﻿using System;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.IO;
using Microsoft.Win32;
using System.Windows.Media.Media3D;

namespace MolekuelViewerHit
{
  public partial class MainWindow : Window
  {
    static int iRinge = 16;
    static int iSeg = 16;

    private AxisAngleRotation3D ax;
    private AxisAngleRotation3D ay;
    private AxisAngleRotation3D az;
    private ScaleTransform3D st;
    private Point3D[,] ptArr = new Point3D[iRinge + 1, iSeg];

    public MainWindow()
    {
      InitializeComponent();
      InitializeModel();
    }

    private void OnEnde(object sender, EventArgs e)
    {
      this.Close();
    }

    private void InitializeModel()
    {
      // Kamera
      PerspectiveCamera camera = new PerspectiveCamera();
      camera.FarPlaneDistance = 100;
      camera.NearPlaneDistance = 1;
      camera.LookDirection = new Vector3D(0, 0, -10);
      camera.UpDirection = new Vector3D(0, 1, 0);
      camera.Position = new Point3D(0, 0, 10);
      camera.FieldOfView = 100;
      this.view.Camera = camera;

      // Licht
      ModelVisual3D lightModel = new ModelVisual3D();
      Vector3D lightVector = new Vector3D(-5, -10, -10);
      DirectionalLight light = new DirectionalLight(Colors.White, lightVector);
      lightModel.Content = light;
      this.view.Children.Add(lightModel);

      // Transformationen definieren
      ax = new AxisAngleRotation3D(new Vector3D(1, 0, 0), 0.0);
      ay = new AxisAngleRotation3D(new Vector3D(0, 1, 0), 0.0);
      az = new AxisAngleRotation3D(new Vector3D(0, 0, 1), 0.0);
      st = new ScaleTransform3D(new Vector3D(1, 1, 1), new Point3D(0, 0, 0));

      // Berechnungen der Normalen und des Einheitskugel
      double dRingWinkel = Math.PI / iRinge;
      double dSegWinkel = 2.0 * Math.PI / iSeg;

      for (int iR = 0; iR <= iRinge; iR++)
      {
        double r0 = Math.Sin(iR * dRingWinkel);
        double y0 = Math.Cos(iR * dRingWinkel);

        for (int iS = 0; iS < iSeg; iS++)
        {
          double x0 = r0 * Math.Sin(iS * dSegWinkel);
          double z0 = r0 * Math.Cos(iS * dSegWinkel);

          ptArr[iR, iS] = new Point3D(x0, y0, z0);
        }
      }
    }

    private void OnLaden(object sender, EventArgs e)
    {
      OpenFileDialog dlg = new OpenFileDialog();

      dlg.Filter = "Molekül-Dateien (*.mol)|*.mol|Alle Dateien (*.*)|*.*";
      dlg.FilterIndex = 0;
      dlg.CheckFileExists = true;

      if (dlg.ShowDialog() == true)
      {
        if (this.view.Children.Count > 1)
        {
          this.view.Children.RemoveAt(1);

          ax.Angle = 0.0;
          ay.Angle = 0.0;
          az.Angle = 0.0;
          st.ScaleX = 1.0;
          st.ScaleY = 1.0;
          st.ScaleZ = 1.0;
        }

        StreamReader sr = new StreamReader(dlg.FileName);

        ModelVisual3D model = new ModelVisual3D();
        Model3DGroup group = new Model3DGroup();

        string strIn = sr.ReadLine();

        while (!string.IsNullOrEmpty(strIn))
        {
          // Farbe, Radius, X, Y, Z
          string[] strData = strIn.Split(',');

          if (strData.Length == 5)
          {
            try
            {
              Color color = GetColor(int.Parse(strData[0]));
              double dRadius = double.Parse(strData[1]);
              Point3D pt = new Point3D(double.Parse(strData[2]), double.Parse(strData[3]),
                                       double.Parse(strData[4]));
              group.Children.Add(ModelKugel(iRinge, iSeg, dRadius, color, pt));
            }
            catch (Exception)
            {
              MessageBox.Show("Fehlerhafte Eingabe!");
            }
          }

          strIn = sr.ReadLine();
        }

        sr.Close();

        // Alle Transformationen in einer Gruppe arrangieren
        Transform3DGroup tg = new Transform3DGroup();
        tg.Children.Add(new RotateTransform3D(ax, new Point3D(0, 0, 0)));
        tg.Children.Add(new RotateTransform3D(ay, new Point3D(0, 0, 0)));
        tg.Children.Add(new RotateTransform3D(az, new Point3D(0, 0, 0)));
        tg.Children.Add(st);
        model.Transform = tg;

        model.Content = group;
        this.view.Children.Add(model);
      }
    }

    private Color GetColor(int iColNum)
    {
      Color col = Colors.Black;

      switch (iColNum)
      {
        case 1:
          col = Colors.Red;
          break;

        case 2:
          col = Colors.Green;
          break;

        case 3:
        default:
          col = Colors.Blue;
          break;

      }

      return col;
    }

    private void OnDrehen(object sender, EventArgs e)
    {
      if (ax == null)
        return;

      Button btn = sender as Button;

      switch (btn.Name)
      {
        case "btnXP":
          ax.Angle += 10.0;
          break;

        case "btnXM":
          ax.Angle -= 10.0;
          break;

        case "btnYP":
          ay.Angle += 10.0;
          break;

        case "btnYM":
          ay.Angle -= 10.0;
          break;

        case "btnZP":
          az.Angle += 10.0;
          break;

        case "btnZM":
          az.Angle -= 10.0;
          break;

      }
    }

    private void OnZoom(object sender, EventArgs e)
    {
      if (st == null)
        return;

      Button btn = sender as Button;

      switch (btn.Name)
      {
        case "btnZoomP":
          st.ScaleX *= 1.1;
          st.ScaleY *= 1.1;
          st.ScaleZ *= 1.1;
          break;

        case "btnZoomM":
          st.ScaleX /= 1.1;
          st.ScaleY /= 1.1;
          st.ScaleZ /= 1.1;
          break;

      }
    }

    private Model3DGroup ModelKugel(int iRinge, int iSeg, double dSize, Color color, Point3D pt)
    {
      Model3DGroup kugel = new Model3DGroup();
      MeshGeometry3D mesh = new MeshGeometry3D();
      int iIndex = 0;
      Point3D pt0, pt1, pt2, pt3;
      Vector3D vec0, vec1, vec2, vec3;

      for (int iR = 0; iR < iRinge; iR++)
      {
        for (int iS = 0; iS < iSeg; iS++)
        {
          pt0 = ptArr[GetR(iR), GetS(iS)];
          vec0 = (Vector3D)pt0;
          pt0.X = pt0.X * dSize + pt.X;
          pt0.Y = pt0.Y * dSize + pt.Y;
          pt0.Z = pt0.Z * dSize + pt.Z;
          pt1 = ptArr[GetR(iR + 1), GetS(iS)];
          vec1 = (Vector3D)pt1;
          pt1.X = pt1.X * dSize + pt.X;
          pt1.Y = pt1.Y * dSize + pt.Y;
          pt1.Z = pt1.Z * dSize + pt.Z;
          pt2 = ptArr[GetR(iR), GetS(iS + 1)];
          vec2 = (Vector3D)pt2;
          pt2.X = pt2.X * dSize + pt.X;
          pt2.Y = pt2.Y * dSize + pt.Y;
          pt2.Z = pt2.Z * dSize + pt.Z;
          pt3 = ptArr[GetR(iR + 1), GetS(iS + 1)];
          vec3 = (Vector3D)pt3;
          pt3.X = pt3.X * dSize + pt.X;
          pt3.Y = pt3.Y * dSize + pt.Y;
          pt3.Z = pt3.Z * dSize + pt.Z;

          mesh.Positions.Add(pt0);
          mesh.Positions.Add(pt1);
          mesh.Positions.Add(pt2);

          mesh.TriangleIndices.Add(iIndex);
          mesh.TriangleIndices.Add(iIndex + 1);
          mesh.TriangleIndices.Add(iIndex + 2);

          mesh.Normals.Add(vec0);
          mesh.Normals.Add(vec1);
          mesh.Normals.Add(vec2);

          mesh.Positions.Add(pt1);
          mesh.Positions.Add(pt3);
          mesh.Positions.Add(pt2);

          mesh.TriangleIndices.Add(iIndex + 3);
          mesh.TriangleIndices.Add(iIndex + 4);
          mesh.TriangleIndices.Add(iIndex + 5);

          mesh.Normals.Add(vec1);
          mesh.Normals.Add(vec3);
          mesh.Normals.Add(vec2);

          iIndex += 6;
        }
      }

      Material mat = new DiffuseMaterial(new SolidColorBrush(color));
      GeometryModel3D model = new GeometryModel3D(mesh, mat);
      kugel.Children.Add(model);

      return kugel;
    }

    private int GetR(int i)
    {
      i = Math.Abs(i);

      if (i > iRinge)
        i = 2 * iRinge - i;

      return i;
    }

    private int GetS(int i)
    {
      if (i < 0)
        i += iSeg;

      return i % iSeg;
    }

    private void OnMouseDown(object sender, MouseButtonEventArgs e)
    {
      // Mausklickposition relativ zum Viewport3D-Element ermitteln
      // Vektor der Suchrichtung vom Klickpunkt in Blickrichtung der Kamera
      PointHitTestParameters phtp = new PointHitTestParameters(e.GetPosition(view));

      // Hittesting durchführen (asynchron)
      VisualTreeHelper.HitTest(view, null, HitResults, phtp);
    }

    public HitTestResultBehavior HitResults(HitTestResult res)
    {
      RayMeshGeometry3DHitTestResult rayMeshRes = res as RayMeshGeometry3DHitTestResult;

      if (rayMeshRes != null)
      {
        // Welches Geometriemodell wurde angeklickt?
        GeometryModel3D hitGeo = rayMeshRes.ModelHit as GeometryModel3D;

        // Ermittlung der Materialfarbe (hier: DiffuseMaterial) 
        DiffuseMaterial mat = hitGeo.Material as DiffuseMaterial;
        SolidColorBrush brush = mat.Brush as SolidColorBrush;

        if (brush.Color.Equals(Colors.Red))
        {
          MessageBox.Show("Rote Kugel gefunden.");
        }
        else if (brush.Color.Equals(Colors.Green))
        {
          MessageBox.Show("Grüne Kugel gefunden.");
        }
        else
        {
          MessageBox.Show("Blaue Kugel");
        }
      }

      // Nur das erste gefundene Objekt ausgeben
      return HitTestResultBehavior.Stop;
    }
  }
}
