﻿using System;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Media3D;
using Microsoft.Win32;

namespace WuerfelBild
{
  public partial class MainWindow : Window
  {
    private AxisAngleRotation3D ax;
    private ScaleTransform3D st;
    private ImageBrush brush;

    public MainWindow()
    {
      InitializeComponent();
    }

    private void OnStartExit(object sender, EventArgs e)
    {
      this.Close();
    }

    private void OnStartWuerfel(object sender, EventArgs e)
    {
      OpenFileDialog dlg = new OpenFileDialog();
      dlg.Filter = "Bild-Dateien (*.jpg)|*.jpg|Bitmap-Dateien (*.bmp)|*.bmp|Alle Dateien (*.*)|*.*";
      dlg.FilterIndex = 0;
      dlg.CheckFileExists = true;

      if (dlg.ShowDialog() != true) return;

      BitmapImage img = new BitmapImage(new Uri(dlg.FileName, UriKind.RelativeOrAbsolute));
      brush = new ImageBrush(img);
      this.view.Children.Clear();

      // Platzierung der Kamera
      PerspectiveCamera camera = new PerspectiveCamera();
      camera.FarPlaneDistance = 100;
      camera.NearPlaneDistance = 1;
      camera.LookDirection = new Vector3D(-10, -10, -7);
      camera.UpDirection = new Vector3D(0, 1, 0);
      camera.Position = new Point3D(10, 10, 10);
      camera.FieldOfView = 70;
      this.view.Camera = camera;

      // Lichteinstellung
      ModelVisual3D lightModel = new ModelVisual3D();
      Vector3D lightVector = new Vector3D(-20, -30, -10);
      DirectionalLight light = new DirectionalLight(Colors.White, lightVector);
      lightModel.Content = light;

      // Modell erzeugen
      Model3DGroup wuerfel = new Model3DGroup();

      // Eckpunkte des Würfels definieren
      Point3D p0 = new Point3D(0, 0, 0);
      Point3D p1 = new Point3D(5, 0, 0);
      Point3D p2 = new Point3D(5, 0, 5);
      Point3D p3 = new Point3D(0, 0, 5);
      Point3D p4 = new Point3D(0, 5, 0);
      Point3D p5 = new Point3D(5, 5, 0);
      Point3D p6 = new Point3D(5, 5, 5);
      Point3D p7 = new Point3D(0, 5, 5);

      wuerfel.Children.Add(RectangleModel(p3, p2, p6, p7)); // Vorderseite
      wuerfel.Children.Add(RectangleModel(p2, p1, p5, p6)); // Rechte Fläche
      wuerfel.Children.Add(RectangleModel(p1, p0, p4, p5)); // Rückseite
      wuerfel.Children.Add(RectangleModel(p0, p3, p7, p4)); // Linke Fläche
      wuerfel.Children.Add(RectangleModel(p7, p6, p5, p4)); // Obere Fläche
      wuerfel.Children.Add(RectangleModel(p2, p3, p0, p1)); // Untere Fläche

      // Transformationen definieren
      ax = new AxisAngleRotation3D(new Vector3D(0, 1, 0), 0.0);
      RotateTransform3D rt = new RotateTransform3D(ax, new Point3D(0, 0, 0));
      st = new ScaleTransform3D(new Vector3D(1, 1, 1), new Point3D(0, 0, 0));

      // Mehrere Transformationen in einer Gruppe arrangieren
      Transform3DGroup tg = new Transform3DGroup();
      tg.Children.Add(rt);
      tg.Children.Add(st);

      if (this.view.Children.Count == 0)
      {
        ModelVisual3D model = new ModelVisual3D();
        model.Content = wuerfel;
        this.view.Children.Add(lightModel);

        // Transformation anwenden
        model.Transform = tg;

        this.view.Children.Add(model);
      }
    }

    private Model3DGroup RectangleModel(Point3D pt0, Point3D pt1, Point3D pt2, Point3D pt3)
    {
      // Definition eines Rechtecks im Drahtmodell
      MeshGeometry3D mesh = new MeshGeometry3D();

      mesh.Positions.Add(pt0);
      mesh.Positions.Add(pt1);
      mesh.Positions.Add(pt2);
      mesh.Positions.Add(pt3);

      // Dreieck Nr. 1
      mesh.TriangleIndices.Add(0);
      mesh.TriangleIndices.Add(1);
      mesh.TriangleIndices.Add(2);

      Vector3D norm = CalcNormal(pt0, pt1, pt2);

      // Normale einfügen
      mesh.Normals.Add(norm);
      mesh.Normals.Add(norm);
      mesh.Normals.Add(norm);

      // Dreieck Nr. 2
      mesh.TriangleIndices.Add(0);
      mesh.TriangleIndices.Add(2);
      mesh.TriangleIndices.Add(3);

      // Gleiche Normale kann hier benutzt werden
      mesh.Normals.Add(norm);
      mesh.Normals.Add(norm);
      mesh.Normals.Add(norm);

      // Verbindungspunkte zum Image definieren
      PointCollection pc = new PointCollection();
      pc.Add(new Point(0, 0));
      pc.Add(new Point(1, 0));
      pc.Add(new Point(1, 1));
      pc.Add(new Point(0, 1));
      mesh.TextureCoordinates = pc;

      // ImageBrush als Material definieren
      Material mat = new DiffuseMaterial(brush);
      GeometryModel3D model = new GeometryModel3D(mesh, mat);
      Model3DGroup group = new Model3DGroup();
      group.Children.Add(model);

      return group;
    }

    private Vector3D CalcNormal(Point3D pt0, Point3D pt1, Point3D pt2)
    {
      // Berechnung des Vektorproduktes
      Vector3D v0 = new Vector3D(pt1.X - pt0.X, pt1.Y - pt0.Y, pt1.Z - pt0.Z);
      Vector3D v1 = new Vector3D(pt2.X - pt1.X, pt2.Y - pt1.Y, pt2.Z - pt1.Z);

      return Vector3D.CrossProduct(v0, v1);
    }

    private void OnStartDrehenSchrittVor(object sender, EventArgs e)
    {
      // Drehen gegen den Uhrzeigersinn
      if (ax != null)
      {
        ax.Angle += 10.0;
      }
    }

    private void OnStartDrehenSchrittRueck(object sender, EventArgs e)
    {
      // Drehen im Uhrzeigersinn
      if (ax != null)
      {
        ax.Angle -= 10.0;
      }
    }

    private void OnStartSchrittGroesser(object sender, EventArgs e)
    {
      // Vergrößeren
      if (st != null)
      {
        st.ScaleX *= 1.1;
        st.ScaleY *= 1.1;
        st.ScaleZ *= 1.1;
      }
    }

    private void OnStartSchrittKleiner(object sender, EventArgs e)
    {
      // Verleinern
      if (st != null)
      {
        st.ScaleX /= 1.1;
        st.ScaleY /= 1.1;
        st.ScaleZ /= 1.1;
      }
    }
  }
}
